"""Pygments lexer for OpenSSL configuration files

With inspiration from IniLexer and BashLexer.
"""

from pygments.lexer import Lexer, LexerContext, RegexLexer, ExtendedRegexLexer, \
    bygroups, include, using, this, do_insertions, default
from pygments.token import Punctuation, Text, Comment, Keyword, Name, String, \
    Generic, Operator, Number, Whitespace, Literal

T_LHS = Name.Attribute
T_RHS = String

# Pygments 2.11 changed the whitespace token type
from pygments import lex, lexers
T_SPACE = list(lex(' ', lexers.get_lexer_by_name('ini')))[0][0]

T_NUMBER = T_RHS
T_EMAIL = T_RHS
T_IP = T_RHS
T_HEX = T_RHS

T_KNOWNDIR = Name.Builtin
T_OTHERDIR = T_LHS

T_KNOWNNAME = Keyword.Pseudo
T_OTHERNAME = T_RHS


class OpenSSLConfLexer(RegexLexer):
    """Pygments lexer for OpenSSL configuration files.
    """

    name = 'OpenSSL'
    aliases = ['openssl']
    filenames = ['*.cnf', '*.conf']
    mimetypes = ['text/x-openssl']

    tokens = {
        'comment': [
            # Comment
            (r'#.*(?=\n)', Comment),
        ],
        'string': [
            # Double-quoted string
            (r'(?s)"(\\\\|\\[0-7]+|\\.|[^"\\])*"', String.Double),
            # Single-quoted string
            (r"(?s)'(\\\\|\\[0-7]+|\\.|[^'\\])*'", String.Single),
        ],
        'variable': [
            # Variable name inside curly braces
            (r'\${', Name.Variable, 'curly-brace'),
            # Variable name inside parentheses
            (r'\$\(', Name.Variable, 'paren'),
            # Variable name
            (r'\$\w+(?:::\w+)?', Name.Variable),
        ],
        'number': [
            # Float
            (r'(?<=\W)\d+\.\d+(?=\W)', T_NUMBER),
            # Int
            (r'(?<=\W)\d+(?=\W)', T_NUMBER),
        ],
        'email': [
            # Email
            (r'[\w\.+-]+\@[\w\.-]+', T_EMAIL),
            default('#pop'),
        ],
        'ip': [
            # IP4
            (r'[\d\.]+', T_IP),
            # IP6
            (r'[\da-fA-F:\.]+', T_IP),
            default('#pop'),
        ],
        'hex': [
            # Hex
            (r'[\da-fA-F:]+', T_HEX),
            default('#pop'),
        ],
        'curly-brace': [
            # Exit condition
            (r'}', Name.Variable, '#pop'),
            # Variable name
            (r'\w+(?:::\w+)?', Name.Variable, 'close-brace'),
            # Whitespace
            (r'\s+', T_SPACE),
            # Catch all
            (r'[^}]+', T_RHS),
        ],
        'close-brace': [
            # Exit condition
            (r'}', Name.Variable, '#pop:2'),
            # Whitespace
            (r'\s+', T_SPACE),
            # Catch all
            (r'[^}]+', T_RHS),
        ],
        'paren': [
            # Exit condition
            (r'\)', Name.Variable, '#pop'),
            # Variable name
            (r'\w+(?:::\w+)?', Name.Variable, 'close-paren'),
            # Whitespace
            (r'\s+', T_SPACE),
            # Catch all
            (r'[^)]+', T_RHS),
        ],
        'close-paren': [
            # Exit condition
            (r'\)', Name.Variable, '#pop:2'),
            # Whitespace
            (r'\s+', T_SPACE),
            # Catch all
            (r'[^)]+', T_RHS),
        ],
        'lhs-default': [
            # Line continuation
            (r'\\(?=\n)', String.Escape),
            # Whitespace
            (r'\s+', T_SPACE),
            # Catch all
            (r'.', T_LHS),
        ],
        'rhs-default': [
            # Line continuation
            (r'\\(?=\n)', String.Escape),
            # Whitespace
            (r'\s+', T_SPACE),
            # Catch all
            (r'.', T_RHS),
        ],
        'root': [
            include('comment'),
            # Section header
            (r'\[.*?\](?=\n)', Keyword),
            # Pragma directive
            (r'(?i)(\.pragma)(?=[\s=\\])([^\S\n]*)(=)([^\S\n]*)',
                bygroups(T_KNOWNDIR, T_SPACE, Operator, T_SPACE), 'pragma'),
            (r'(?i)(\.pragma)(?=[\s\\])([^\S\n]*)',
                bygroups(T_KNOWNDIR, T_SPACE), 'pragma'),
            # Include directive
            (r'(?i)(\.include)(?=[\s=\\])([^\S\n]*)(=)([^\S\n]*)',
                bygroups(T_KNOWNDIR, T_SPACE, Operator, T_SPACE), 'other'),
            (r'(?i)(\.include)(?=[\s\\])([^\S\n]*)',
                bygroups(T_KNOWNDIR, T_SPACE), 'other'),
            # Other directives
            (r'(\.[\w-]+)([^\S\n]*)(=)([^\S\n]*)',
                bygroups(T_OTHERDIR, T_SPACE, Operator, T_SPACE), 'other'),
            (r'(\.[\w-]+)([^\S\n]*)',
                bygroups(T_OTHERDIR, T_SPACE), 'other'),
            # Left hand side
            (r'([\w\.;-]+)(\s*)', bygroups(T_LHS, T_SPACE)),
            # Operator
            (r'(=)([^\S\n]*)', bygroups(Operator, T_SPACE), 'rhs'),
            include('lhs-default'),
        ],
        'rhs': [
            include('comment'),
            # Exit condition
            (r'(?<!\\)\n', T_SPACE, '#pop'),
            # Email
            (r'(?i)(?<=\W)(email)(?=\W)([^\S\n]*)(:)([^\S\n]*)',
                bygroups(T_RHS, T_SPACE, T_RHS, T_SPACE), 'email'),
            # IP
            (r'(?i)(?<=\W)(IP)(?=\W)([^\S\n]*)(:)([^\S\n]*)',
                bygroups(T_RHS, T_SPACE, T_RHS, T_SPACE), 'ip'),
            # DER
            (r'(?i)(?<=\W)(DER)(?=\W)([^\S\n]*)(:)([^\S\n]*)',
                bygroups(T_RHS, T_SPACE, T_RHS, T_SPACE), 'hex'),
            include('string'),
            include('variable'),
            # OID
            (r'(?<=\W)\d+\.(?:\d+\.?)*(?=\W)', Name.Function),
            include('number'),
            # Section reference
            (r'(?<=\W)\@\w+', Name.Constant),
            # Critical keyword
            (r'(?i)(?<=\W)critical(?=\W)', Keyword.Pseudo),
            include('rhs-default'),
        ],
        'pragma': [
            include('comment'),
            # Exit condition
            (r'(?<!\\)\n', T_SPACE, '#pop'),
            # Known directive names
            (r'(?i)(?<=\W)(abspath|dollarid|includedir)(?=\W)([^\S\n]*)(:)([^\S\n]*)',
                bygroups(T_KNOWNNAME, T_SPACE, Operator, T_SPACE), 'value'),
            # Other directive names
            (r'([\w-]+)([^\S\n]*)(:)([^\S\n]*)',
                bygroups(T_OTHERNAME, T_SPACE, T_RHS, T_SPACE), 'value'),
            include('string'),
            include('variable'),
            include('number'),
            include('rhs-default'),
        ],
        'other': [
            include('comment'),
            # Exit condition
            (r'(?<!\\)\n', T_SPACE, '#pop'),
            include('string'),
            include('variable'),
            include('number'),
            include('rhs-default'),
        ],
        'value': [
            include('comment'),
            # Exit condition
            (r'(?<!\\)\n', T_SPACE, '#pop:2'),
            include('string'),
            include('variable'),
            include('number'),
            include('rhs-default'),
        ],
    }

    def __init__(self, **options):
        super(OpenSSLConfLexer, self).__init__(**options)
        # Always apply tokenmerge filter
        self.add_filter('tokenmerge')

    def analyse_text(text):
        npos = text.find('\n')
        if npos < 3:
            return False
        return text[0] == '[' and text[npos-1] == ']'



import unittest


class InstallationTests(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_has_lexer(self):
        from pygments.lexers import get_lexer_by_name
        get_lexer_by_name('openssl')

    def test_unknown_lexer_raises(self):
        from pygments.lexers import get_lexer_by_name
        from pygments.util import ClassNotFound
        try:
            get_lexer_by_name('does-not-exist')
        except ClassNotFound:
            pass # success
        else:
            self.fail('ClassNotFound not raised')




import unittest

from pygments_openssl.lexer import T_SPACE


class LexerTests(unittest.TestCase):

    def lex(self, code, lexer_name):
        from pygments import lex, lexers
        return list(lex(code, lexers.get_lexer_by_name(lexer_name)))

    def test_lex_comment(self):
        from pygments import token

        tokens = self.lex('# Comment\n', 'openssl')
        self.assertEqual(tokens[0], (token.Comment, '# Comment'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))

        tokens = self.lex('# Comment\n', 'ini')
        self.assertEqual(tokens[0], (token.Comment.Single, '# Comment'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))

        tokens = self.lex('# Comment\n', 'bash')
        self.assertEqual(tokens[0], (token.Comment.Single, '# Comment\n'))

    def test_lex_section_header(self):
        from pygments import token

        tokens = self.lex('[ default ]\n', 'openssl')
        self.assertEqual(tokens[0], (token.Keyword, '[ default ]'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))

        tokens = self.lex('[ default ]\n', 'ini')
        self.assertEqual(tokens[0], (token.Keyword, '[ default ]'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))

    def test_lex_lhs_and_operator(self):
        from pygments import token

        tokens = self.lex('dir = .\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))

        tokens = self.lex('dir = .\n', 'ini')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))

    def test_lex_lhs_line_continuation(self):
        from pygments import token

        tokens = self.lex('dir \\\n = .\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.String.Escape, '\\'))
        self.assertEqual(tokens[3], (T_SPACE, '\n '))
        self.assertEqual(tokens[4], (token.Operator, '='))
        self.assertEqual(tokens[5], (T_SPACE, ' '))

    def test_lex_rhs_line_continuation(self):
        from pygments import token

        tokens = self.lex('dir = \\\n.\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String.Escape, '\\'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_rhs_string(self):
        from pygments import token

        tokens = self.lex('dir = .\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, '.'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

        tokens = self.lex('dir = .\n', 'ini')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, '.'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_rhs_comment(self):
        from pygments import token

        tokens = self.lex('dir = . # Comment\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, '.'))
        self.assertEqual(tokens[5], (T_SPACE, ' '))
        self.assertEqual(tokens[6], (token.Comment, '# Comment'))
        self.assertEqual(tokens[7], (T_SPACE, '\n'))

        tokens = self.lex('dir = . # Comment\n', 'ini')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, '. # Comment'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_rhs_double_quoted_string(self):
        from pygments import token

        tokens = self.lex('dir = "foo bar"\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String.Double, '"foo bar"'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

        tokens = self.lex('dir = "foo bar"\n', 'ini')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, '"foo bar"'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_rhs_single_quoted_string(self):
        from pygments import token

        tokens = self.lex("dir = 'foo bar'\n", 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String.Single, "'foo bar'"))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

        tokens = self.lex("dir = 'foo bar'\n", 'ini')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, "'foo bar'"))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_rhs_variable_name(self):
        from pygments import token

        tokens = self.lex('foo = $variable\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'foo'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.Name.Variable, '$variable'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_rhs_variable_name_curly_braces(self):
        from pygments import token

        tokens = self.lex('foo = ${ENV::variable}\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'foo'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.Name.Variable, '${ENV::variable}'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_rhs_variable_name_parentheses(self):
        from pygments import token

        tokens = self.lex('foo = $(ENV::variable)\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'foo'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.Name.Variable, '$(ENV::variable)'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_rhs_oid(self):
        from pygments import token

        tokens = self.lex('oid = 1.2.3.4.5\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'oid'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.Name.Function, '1.2.3.4.5'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_rhs_number(self):
        from pygments import token

        tokens = self.lex('num = 12\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'num'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, '12'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_rhs_section_reference(self):
        from pygments import token

        tokens = self.lex('foo = @section\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'foo'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.Name.Constant, '@section'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_rhs_critical_keyword(self):
        from pygments import token

        tokens = self.lex('foo = critical,bar\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'foo'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.Keyword.Pseudo, 'critical'))
        self.assertEqual(tokens[5], (token.String, ',bar'))
        self.assertEqual(tokens[6], (T_SPACE, '\n'))

    def test_lex_incomplete_lhs(self):
        from pygments import token

        tokens = self.lex('dir\ndir = .\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))
        self.assertEqual(tokens[2], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.Operator, '='))
        self.assertEqual(tokens[5], (T_SPACE, ' '))
        self.assertEqual(tokens[6], (token.String, '.'))
        self.assertEqual(tokens[7], (T_SPACE, '\n'))

    def test_lex_incomplete_lhs_and_operator(self):
        from pygments import token

        tokens = self.lex('dir =\ndir = .\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, '\n'))
        self.assertEqual(tokens[4], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[5], (T_SPACE, ' '))
        self.assertEqual(tokens[6], (token.Operator, '='))
        self.assertEqual(tokens[7], (T_SPACE, ' '))
        self.assertEqual(tokens[8], (token.String, '.'))
        self.assertEqual(tokens[9], (T_SPACE, '\n'))

    def test_lex_incomplete_lhs_string(self):
        from pygments import token

        tokens = self.lex('dir', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, 'dir'))

    def test_lex_missing_lhs(self):
        from pygments import token

        tokens = self.lex('= foo\ndir = .\n', 'openssl')
        self.assertEqual(tokens[0], (token.Operator, '='))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.String, 'foo'))
        self.assertEqual(tokens[3], (T_SPACE, '\n'))
        self.assertEqual(tokens[4], (token.Name.Attribute, 'dir'))
        self.assertEqual(tokens[5], (T_SPACE, ' '))
        self.assertEqual(tokens[6], (token.Operator, '='))
        self.assertEqual(tokens[7], (T_SPACE, ' '))
        self.assertEqual(tokens[8], (token.String, '.'))
        self.assertEqual(tokens[9], (T_SPACE, '\n'))


class DirectiveLexerTests(unittest.TestCase):

    def lex(self, code, lexer_name):
        from pygments import lex, lexers
        return list(lex(code, lexers.get_lexer_by_name(lexer_name)))

    def test_lex_directive(self):
        from pygments import token

        tokens = self.lex('.directive foo\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, '.directive'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.String, 'foo'))
        self.assertEqual(tokens[3], (T_SPACE, '\n'))

    def test_lex_directive_and_operator(self):
        from pygments import token

        tokens = self.lex('.directive = foo\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, '.directive'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, 'foo'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_directive_with_leading_whitespace(self):
        from pygments import token

        tokens = self.lex('  .directive foo\n', 'openssl')
        self.assertEqual(tokens[0], (T_SPACE, '  '))
        self.assertEqual(tokens[1], (token.Name.Attribute, '.directive'))
        self.assertEqual(tokens[2], (T_SPACE, ' '))
        self.assertEqual(tokens[3], (token.String, 'foo'))
        self.assertEqual(tokens[4], (T_SPACE, '\n'))

    def test_lex_incomplete_directive(self):
        from pygments import token

        tokens = self.lex('.directive\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, '.directive'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))

        tokens = self.lex('.directive\n.directive foo\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, '.directive'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))
        self.assertEqual(tokens[2], (token.Name.Attribute, '.directive'))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, 'foo'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_incomplete_directive_and_operator(self):
        from pygments import token

        tokens = self.lex('.directive =\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, '.directive'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, '\n'))

        tokens = self.lex('.directive =\n.directive = foo\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, '.directive'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, '\n'))
        self.assertEqual(tokens[4], (token.Name.Attribute, '.directive'))
        self.assertEqual(tokens[5], (T_SPACE, ' '))
        self.assertEqual(tokens[6], (token.Operator, '='))
        self.assertEqual(tokens[7], (T_SPACE, ' '))
        self.assertEqual(tokens[8], (token.String, 'foo'))
        self.assertEqual(tokens[9], (T_SPACE, '\n'))

    def test_lex_incomplete_directive_string(self):
        from pygments import token

        tokens = self.lex('.directive', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, '.directive'))

        tokens = self.lex('.directive\n.directive foo', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Attribute, '.directive'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))
        self.assertEqual(tokens[2], (token.Name.Attribute, '.directive'))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, 'foo'))


class PragmaDirectiveLexerTests(unittest.TestCase):

    def lex(self, code, lexer_name):
        from pygments import lex, lexers
        return list(lex(code, lexers.get_lexer_by_name(lexer_name)))

    def test_lex_pragma_directive(self):
        from pygments import token

        tokens = self.lex('.pragma foo\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.String, 'foo'))
        self.assertEqual(tokens[3], (T_SPACE, '\n'))

    def test_lex_pragma_directive_and_operator(self):
        from pygments import token

        tokens = self.lex('.pragma = foo\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, 'foo'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_pragme_directive_with_leading_whitespace(self):
        from pygments import token

        tokens = self.lex('  .pragma foo\n', 'openssl')
        self.assertEqual(tokens[0], (T_SPACE, '  '))
        self.assertEqual(tokens[1], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[2], (T_SPACE, ' '))
        self.assertEqual(tokens[3], (token.String, 'foo'))
        self.assertEqual(tokens[4], (T_SPACE, '\n'))

    def test_lex_pragma_directive_name(self):
        from pygments import token

        tokens = self.lex('.pragma abspath:\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Keyword.Pseudo, 'abspath'))
        self.assertEqual(tokens[3], (token.Operator, ':'))

    def test_lex_pragma_directive_name_and_operator(self):
        from pygments import token

        tokens = self.lex('.pragma = abspath:\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.Keyword.Pseudo, 'abspath'))
        self.assertEqual(tokens[5], (token.Operator, ':'))

    def test_lex_pragma_directive_name_and_value(self):
        from pygments import token

        tokens = self.lex('.pragma abspath:bar\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Keyword.Pseudo, 'abspath'))
        self.assertEqual(tokens[3], (token.Operator, ':'))
        self.assertEqual(tokens[4], (token.String, 'bar'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_pragma_directive_name_and_value_with_colon(self):
        from pygments import token

        tokens = self.lex('.pragma abspath:bar:baz\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Keyword.Pseudo, 'abspath'))
        self.assertEqual(tokens[3], (token.Operator, ':'))
        self.assertEqual(tokens[4], (token.String, 'bar:baz'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_incomplete_pragma_directive(self):
        from pygments import token

        tokens = self.lex('.pragma\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))

        tokens = self.lex('.pragma\n.pragma foo\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))
        self.assertEqual(tokens[2], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, 'foo'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_incomplete_pragma_directive_and_operator(self):
        from pygments import token

        tokens = self.lex('.pragma =\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, '\n'))

        tokens = self.lex('.pragma =\n.pragma = foo\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, '\n'))
        self.assertEqual(tokens[4], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[5], (T_SPACE, ' '))
        self.assertEqual(tokens[6], (token.Operator, '='))
        self.assertEqual(tokens[7], (T_SPACE, ' '))
        self.assertEqual(tokens[8], (token.String, 'foo'))
        self.assertEqual(tokens[9], (T_SPACE, '\n'))

    def test_lex_incomplete_pragma_directive_string(self):
        from pygments import token

        tokens = self.lex('.pragma', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.pragma'))

        tokens = self.lex('.pragma\n.pragma foo', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))
        self.assertEqual(tokens[2], (token.Name.Builtin, '.pragma'))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, 'foo'))


class IncludeDirectiveLexerTests(unittest.TestCase):

    def lex(self, code, lexer_name):
        from pygments import lex, lexers
        return list(lex(code, lexers.get_lexer_by_name(lexer_name)))

    def test_lex_include_directive(self):
        from pygments import token

        tokens = self.lex('.include foo\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.include'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.String, 'foo'))
        self.assertEqual(tokens[3], (T_SPACE, '\n'))

    def test_lex_include_directive_and_operator(self):
        from pygments import token

        tokens = self.lex('.include = foo\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.include'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, 'foo'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_pragme_directive_with_leading_whitespace(self):
        from pygments import token

        tokens = self.lex('  .include foo\n', 'openssl')
        self.assertEqual(tokens[0], (T_SPACE, '  '))
        self.assertEqual(tokens[1], (token.Name.Builtin, '.include'))
        self.assertEqual(tokens[2], (T_SPACE, ' '))
        self.assertEqual(tokens[3], (token.String, 'foo'))
        self.assertEqual(tokens[4], (T_SPACE, '\n'))

    def test_lex_incomplete_include_directive(self):
        from pygments import token

        tokens = self.lex('.include\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.include'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))

        tokens = self.lex('.include\n.include foo\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.include'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))
        self.assertEqual(tokens[2], (token.Name.Builtin, '.include'))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, 'foo'))
        self.assertEqual(tokens[5], (T_SPACE, '\n'))

    def test_lex_incomplete_include_directive_and_operator(self):
        from pygments import token

        tokens = self.lex('.include =\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.include'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, '\n'))

        tokens = self.lex('.include =\n.include = foo\n', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.include'))
        self.assertEqual(tokens[1], (T_SPACE, ' '))
        self.assertEqual(tokens[2], (token.Operator, '='))
        self.assertEqual(tokens[3], (T_SPACE, '\n'))
        self.assertEqual(tokens[4], (token.Name.Builtin, '.include'))
        self.assertEqual(tokens[5], (T_SPACE, ' '))
        self.assertEqual(tokens[6], (token.Operator, '='))
        self.assertEqual(tokens[7], (T_SPACE, ' '))
        self.assertEqual(tokens[8], (token.String, 'foo'))
        self.assertEqual(tokens[9], (T_SPACE, '\n'))

    def test_lex_incomplete_include_directive_string(self):
        from pygments import token

        tokens = self.lex('.include', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.include'))

        tokens = self.lex('.include\n.include foo', 'openssl')
        self.assertEqual(tokens[0], (token.Name.Builtin, '.include'))
        self.assertEqual(tokens[1], (T_SPACE, '\n'))
        self.assertEqual(tokens[2], (token.Name.Builtin, '.include'))
        self.assertEqual(tokens[3], (T_SPACE, ' '))
        self.assertEqual(tokens[4], (token.String, 'foo'))

